class AlvPossibleFormatError(Exception):
    '''
    When reading a file yields no records, so probably wrong format.
    '''
    def __init__(self, expression, filename):
        self.expression = 'HUBBA'
        self.message = 'No sequence records in ' + filename



from alv.get_terminal_size import get_terminal_size

class AlignmentTerminal:
    '''
    This class encapsulates knowledge about the terminal and how to draw on it.
    '''
    def __init__(self, args):
        self.width = get_terminal_size()[0] # First item in this tuple is the width, the other is the height.
        self.sorting = args.sorting
        if args.sort_by_id:
            self.sorting = 'by identity'
            self.order = args.sort_by_id
        elif args.sorting_order:
            self.sorting = 'fixed'
            self.order = args.sorting_order.split(',')
            if len(self.order) == 0:
                raise Exception('Bad order specification: no accessions in input')

    def output_alignment(self, al, painter, args):
        '''
        Output alignment al to stdout in blocks of width at most w with colors from painter.
        '''
        self.left_margin = 1 + al.accession_widths()
        assert self.left_margin < self.width - 10

        if self.sorting == 'alpha':
            accessions = al.sorted_accessions()
        elif self.sorting == 'fixed':
            accessions = self.order
        elif self.sorting == 'by identity':
            accessions = al.sort_by_identity(self.order)
        else:
            accessions = al.accessions()
        accessions = list(accessions)
            
        columns_per_block = al.block_width(self.width, args)
        for block in al.blocks(columns_per_block, args):
            for acc in accessions:
                colored_subseq = al.apply_painter(acc, block, painter)
                print("{0:{width}}{1}".format(acc, colored_subseq, width=self.left_margin))
            print(' ' * self.left_margin, block.start, sep='')


    def print_one_sequence_block(self, record, left_margin, start, block_width):
        colored_string = colorize_sequence_string(rec.seq[start : start + block_width])
        print("{0:{width}}{1}".format(rec.id, colored_string, width=left_margin))


    #
    # Needs reimplementation: Loop over columns until a block is full
    # 

    # for start in range(0, al_width, block_width):
    #     for record in alignment:
    #         painter.print_one_sequence_block(record, left_margin, start, block_width)
    #     print(' ' * left_margin, start, sep='')

        



import unittest
import argparse
from Bio.Alphabet import generic_dna, generic_protein
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment
import alv.io
from alv.alignment import aaAlignment, dnaAlignment, codonAlignment, BaseAlignment
import alv.colorize
import alv.exceptions

class TestIO_etc(unittest.TestCase):
    def test_guess_seq_type(self):
        
        coding = MultipleSeqAlignment([
             SeqRecord(Seq("ACTGCTAGCTAG", generic_dna), id="Alpha"),
             SeqRecord(Seq("ACT-CTAGCTAG", generic_dna), id="Beta"),
             SeqRecord(Seq("ACTGCTAGDTAG", generic_dna), id="Gamma"),
         ])
        noncoding = MultipleSeqAlignment([
             SeqRecord(Seq("CGTCGTCGTCGT", generic_dna), id="Alpha"),
             SeqRecord(Seq("CGACGACGACGA", generic_dna), id="Beta"),
             SeqRecord(Seq("ATAATAATAATA", generic_dna), id="Gamma"),
         ])
        peptide = MultipleSeqAlignment([
             SeqRecord(Seq("ARNDCEQHIKL*", generic_protein), id="Alpha"),
             SeqRecord(Seq("ARNDCEQHIKL*", generic_protein), id="Beta"),
             SeqRecord(Seq("ARNDCEQHIKLM", generic_protein), id="Gamma"),
         ])
        self.assertEqual(alv.io.guess_seq_type(coding), 'codon')
        self.assertEqual(alv.io.guess_seq_type(noncoding), 'dna')
        self.assertEqual(alv.io.guess_seq_type(peptide), 'aa')


class TestFormats(unittest.TestCase):
    def setUp(self):
        self.dna_filename = 'tests/t1.fa'
        self.aa_filename = 'tests/t4.fa'
        self.sthlm_filename= 'tests/t4.sthlm'
        self.pfam_file = 'tests/PF00005_seed.txt'

    def test_reading_fasta_files(self):
        al, painter = alv.io.read_alignment(self.dna_filename, 'dna', 'fasta', '', 'standard')
        self.assertIsInstance(al, dnaAlignment)
        self.assertIsInstance(painter, alv.colorize.dnaPainter)

        al, painter = alv.io.read_alignment(self.dna_filename, 'aa', 'fasta', 'clustal', 'standard')
        self.assertIsInstance(al, aaAlignment)
        self.assertIsInstance(painter, alv.colorize.aaPainter)

        al, painter = alv.io.read_alignment(self.dna_filename, 'codon', 'fasta', 'clustal', 'standard')
        self.assertIsInstance(al, codonAlignment)
        self.assertIsInstance(painter, alv.colorize.codonPainter)

    def test_reading_stockholm_files(self):
        al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'stockholm', '', 'standard')
        self.assertIsInstance(al, aaAlignment)
        self.assertIsInstance(painter, alv.colorize.aaPainter)
        al, painter = alv.io.read_alignment(self.pfam_file, 'aa', 'stockholm', '', 'standard')
        self.assertIsInstance(al, aaAlignment)
        self.assertIsInstance(painter, alv.colorize.aaPainter)


    def test_reading_wrong_format(self):
        with self.assertRaises(alv.exceptions.AlvPossibleFormatError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'clustal', '', 'standard')
        with self.assertRaises(alv.exceptions.AlvPossibleFormatError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'phylip', '', 'standard')
        with self.assertRaises(alv.exceptions.AlvPossibleFormatError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'stockholm', '', 'standard')
        with self.assertRaises(alv.exceptions.AlvPossibleFormatError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'fasta', '', 'standard')
        with self.assertRaises(alv.exceptions.AlvPossibleFormatError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'clustal', '', 'standard')
        with self.assertRaises(alv.exceptions.AlvPossibleFormatError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'phylip', '', 'standard')
        

    def test_guessing_seq_type(self):
        '''
        Given simple DNA or AA input, is the sequence typ (DNA/AA) guessed correctly?
        '''
        al, painter = alv.io.read_alignment(self.dna_filename, 'guess', 'fasta', '', 'standard')
        self.assertIsInstance(al, dnaAlignment)
        al, painter = alv.io.read_alignment(self.aa_filename, 'guess', 'fasta', '', 'standard')
        self.assertIsInstance(al, aaAlignment)
        
if __name__ == '__main__':
    unittest.main()
    



#! /usr/bin/env python
from Bio import AlignIO
from colorama import init as colorama_init, Fore, Back, Style
from get_terminal_size import get_terminal_size
import argparse
import sys
 
def colorizer(c):
    if c in 'AILMFWVC':
        return Back.BLUE + c
    elif c in 'KR':
        return Back.RED + c
    elif c in 'ED':
        return Back.MAGENTA + c
    elif c in 'NQST':
        return Back.GREEN + c
    elif c in 'G':
        return Back.YELLOW + c
    elif c in 'P':
        return Back.YELLOW + c
    elif c in 'HY':
        return Back.CYAN + c
    else:
        return Back.WHITE + c

def colorize_sequence_string(s):
    colored_s = ''
    for c in s:
        colored_s += colorizer(c)
    return colored_s + Style.RESET_ALL

def get_accession_widths(alignment):
    '''
    Compute the space needed for all accessions, plus one for
    a delimiter.
    '''
    max_accession_length = 5        # initial guess, or minimum
    for record in alignment:
        if len(record.id) + 1 > max_accession_length:
            max_accession_length = len(record.id) + 1
    return max_accession_length

def print_one_sequence_block(rec, left_margin, start, block_width):
    colored_string = colorize_sequence_string(rec.seq[start : start + block_width])
    print("{0:{width}}{1}".format(rec.id, colored_string, width=left_margin))

def get_block_width(al_width, left_margin):
    '''
    We want to break a long alignent up into blocks. If so, how big blocks?
    Take the margin size (for accessions) into account and avoid ending up
    with blocks of size 1.
    '''
    terminal_width = get_terminal_size()[0]
    last_block_width = al_width % (terminal_width - left_margin)
    n_blocks = al_width // (terminal_width - left_margin) # Not counting last block, if uneven

    if last_block_width == al_width:
        # Short alignment, fits in one block
        return al_width
    elif last_block_width > 10:
        return terminal_width - left_margin
    else:
        sacrifice = max([1, (10 - last_block_width) // n_blocks])
        print('sacrifice:', sacrifice)
        return terminal_width - left_margin - sacrifice

def main():
    ap = argparse.ArgumentParser()
    ap.add_argument('infile')
    ap.add_argument('-w', '--width', type=int, default=0, help='Width of alignment blocks. Defaults to terminal width minus accession width, essentially.')
    ap.add_argument('-k', '--keep-colors-when-redirecting', action='store_true', help='Do not strip colors when redirecting to stdout, or similar.')
#    ap.add_argument('-p', '--prefix', type=int, default=0, help='Number of characters to remove from the beggining of the accession. Note: PHYLIP format allows max 10 character accessions.')
    args = ap.parse_args()

    alignment = AlignIO.read(args.infile, "fasta")
    al_width = alignment.get_alignment_length()
    left_margin = get_accession_widths(alignment)
    if args.width == 0:
        block_width = get_block_width(al_width, left_margin)
    else:
        block_width = args.width

    # Set up colors
    if args.keep_colors_when_redirecting:
        colorama_init(strip=False)
    else:
        colorama_init()

    for start in range(0, al_width, block_width):
        for record in alignment:
            print_one_sequence_block(record, left_margin, start, block_width)
        print(' ' * left_margin, start, sep='')

if __name__ == '__main__':
    main()
    
